package lsedit;

import java.awt.Container;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Event;
import java.awt.Graphics;
import java.awt.Insets;
import java.awt.Rectangle;

import java.awt.event.*;
import java.net.*;
import java.io.*;
import java.util.*;
import javax.swing.*;

abstract public class LandscapeEditorCore extends LandscapeViewerCore /* extends JPanel */ implements ItemListener 
{
	protected final static int GRID_MAX			 = 10;

	protected final static int NUM_TOOL_BUTTONS	 = 10;

	protected final static String TITLE = "Landscape Editor";

	protected final static int FA_SAVE		 = FA_LOAD+1;
	protected final static int FA_PRINT		 = FA_LOAD+2;
	protected final static int FA_PRINT_ALL	 = FA_LOAD+3;
	protected final static int FA_PRINT_BOOK = FA_LOAD+4;

	protected EditModeHandler m_editModeHandler;
	protected ViewModeHandler m_viewModeHandler;

	protected Vector diagrams = new Vector();
	protected Vector diagramNames = new Vector();

	/* Parameters */

	protected String lsSavePath, lsSaveSuffix, lsSaveCmd;

	protected String printPrompt(String banner, String prompt, String file, int mode)
	{
		JFrame f = Util.getFrame(this);

		PrintPromptDialog.Create(this, f, banner, prompt, file, mode);

		return null;
	}

	// Driving component of quick sort

	private static final int partition( Vector v, int l, int r, int mode ) 
	{
		// Arbitrarily pick the left element as the pivot element:

		Rectangle lytp = ((EntityInstance) v.elementAt(l)).getDiagramBounds();

		l--; 
		r++;

		while(true) {
			// Figure out what's before and after the pivot:

			boolean f;
			do {
				r--;

				Rectangle lyt = ((EntityInstance) v.elementAt(r)).getDiagramBounds();

				if (mode == Do.SPC_HORIZ) {
					f = lyt.x > lytp.x;
				} else {
					f = lyt.y > lytp.y;
				}
			} while (f);

			do {
				l++;
				Rectangle lyt = ((EntityInstance) v.elementAt(l)).getDiagramBounds();
				if (mode == Do.SPC_HORIZ) {
					f = lyt.x < lytp.x;
				} else {	
					f = lyt.y < lytp.y;
				}
			} while (f);
		
			// Swap elements if we can:

			if (l < r) {
				Object o = v.elementAt(l);
				v.setElementAt(v.elementAt(r), l);
				v.setElementAt(o, r);
			} else {
				return r;
	}	}	}

	private static final void sort( Vector v, int l, int r, int mode ) 
	{
		// If we haven't reached a termination condition...

		if (l < r) {
			//	Partition the vector into left and right halves:
			int p = partition(v, l, r, mode);

			//	Recursively sort each half:
			sort(v, l,   p, mode);
			sort(v, p+1, r, mode);
		}
	}

	protected static void sortVectorByLayout(Vector v, int mode) 
	{
		sort(v, 0, v.size() - 1, mode);
	}

	protected String getTitle() 
	{
		return TITLE;
	}

	protected void genDropDowns() 
	{
		Dimension	dim;

		dim = new Dimension(100, ToolBarButton.HEIGHT);
		m_lsDropDown = new JComboBox();
		m_lsDropDown.setMaximumSize(dim);
		m_lsDropDown.setMinimumSize(dim);
		m_lsDropDown.setPreferredSize(dim);
		m_lsDropDown.setSize(dim);
		m_lsDropDown.addItemListener(this);
	}

	protected void genModeHandlers() 
	{
		m_editModeHandler = new EditModeHandler();
		m_editModeHandler.init(this);

		m_viewModeHandler = new ViewModeHandler();
		m_viewModeHandler.init(this);
		m_modeHandler = m_editModeHandler;
	}

	protected int getNumToolButtons() 
	{
		return toolButton.length;
	}

	protected void registerLsName(String name) 
	{
		if (m_lsDropDown != null) {
			m_lsDropDown.addItem(name);

			if (m_lsDropDown.getItemCount() > 1) {
				m_lsDropDown.enable();
	}	}	}


	protected void setDiagram(Diagram value) 
	{
		String	name, diagramName;
		int		i;

		super.setDiagram(value);

		if (m_diagram != null) {
			m_diagram.setDrawBends(!(m_modeHandler instanceof ViewModeHandler));
			diagramName = m_diagram.getName();
			if (diagramName != null) {
				for (i = 0; i < m_lsDropDown.getItemCount(); ++i) {
					name = (String) m_lsDropDown.getItemAt(i);
					if (diagramName.equals(name)) {
						m_lsDropDown.setSelectedIndex(i);
						break;
		}	}	}	}

		if (m_diagram != null) {
			doFeedback("Set to: " + m_diagram.getName());
		}
		setLeftBox();
	}

	protected void setDiagram(int ind) 
	{
		Diagram newDg = (Diagram) diagrams.elementAt(ind);

		if (newDg != m_diagram) {
			setDiagram(newDg);
	}	}

	protected void addDiagram(Diagram newDg, String path) 
	{
		diagrams.addElement(newDg);
		diagramNames.addElement(path);

		registerLsName(newDg.getName());

		setDiagram(newDg);
		newDg.setToViewport();
	}

	public void attach(String lsPath) 
	{
		int		ind;
		Diagram	dgLink;

		ind = diagramNames.indexOf(lsPath);
		if (ind < 0) {
			dgLink = new Diagram(this, true);

			String rc = dgLink.loadDiagram(lsPath, null, null);
		  
			if (rc != null) {
				error("Attach failed: " + rc);
				return;
			}
			dgLink.prepostorder();
			addDiagram(dgLink, lsPath);
		}
		else {
			setDiagram(ind);
		}
	}

	public void followLink(String url, int target) 
	{
		if (target == LsLink.TARGET_APP)
			attach(url);
		else
			showURL(url, target);
	}

	protected void loadLs(String file) 
	{
		if (file != null && file.length() > 0) {
			lsPath = file;

			doFeedback("Reading: " + lsPath);

			Diagram newDg = new Diagram(this, true);

			String rc = newDg.loadDiagram(lsPath, null, null);

			if (rc != null) {
				error("Failed to load (" + rc + ")");
			} else {
				newDg.prepostorder();
				addDiagram(newDg, lsPath);
			}
		}

		this.requestFocus();
		repaint();
	}

	protected void fitChildren() 
	{
		m_diagram.saveForUndo();
		m_diagram.fitTo(m_diagram.getDrawRoot());
		doFeedback("Children fitted to: " + m_diagram.getDrawRoot().getLabel());
		m_diagram.rescaleDiagram();
		repaintDg();
	}

	protected void fitToLabel() 
	{
		Vector v = startGroupOp();

		if (v == null)
			return;
		
		int H_MARGIN = 8; 
		int V_MARGIN = 4;

		Enumeration en;

		for (en = v.elements(); en.hasMoreElements(); ) {
			EntityInstance e = (EntityInstance) en.nextElement();

			Dimension ld = e.getLabelDim(getGraphics(), EntityInstance.REG_FONT);

			Rectangle curLayout = e.getDiagramBounds();

			curLayout.width  = ld.width+H_MARGIN*2;
			curLayout.height = ld.height+V_MARGIN*2;

			m_diagram.saveForUndo();
			e.setDiagramBounds(curLayout);
		}

		m_diagram.rescaleDiagram();
		redrawDg();
	}

	protected Vector startGroupOp() 
	{
		m_diagram.saveForUndo();

		Vector grp = m_diagram.getGroup();

		if (grp == null) {
			error("Group not selected");
			return null;
		}
		return grp;
	}

	// Alignment

	protected void align(int alignment) 
	{
		Vector grp = startGroupOp();

		if (grp == null) {
			return;
		}

		// Align group based on alignment to key entity

		EntityInstance ke = m_diagram.getKeyEntity();
		Rectangle klyt = ke.getDiagramBounds();

		int cx = klyt.x + klyt.width/2;
		int cy = klyt.y + klyt.height/2;

		Enumeration en;

		for (en = grp.elements(); en.hasMoreElements(); ) {
			EntityInstance e = (EntityInstance) en.nextElement();

			if (e != ke) {
				Rectangle elyt = (Rectangle) e.getDiagramBounds().clone();

				switch(alignment)
				{
				case /* ALT */ Do.A_HORIZ_TOP:
					elyt.y = klyt.y;
					break;

				case /* ALT */ Do.A_HORIZ_CENTER:
					elyt.y = cy - elyt.height/2;
					break;

				case /* ALT */ Do.A_VERTICAL_LEFT:
					elyt.x = klyt.x;
					break;

				case /* ALT */ Do.A_VERTICAL_RIGHT:
					elyt.x += klyt.x + klyt.width - elyt.width;
					break; 

				case /* ALT */ Do.A_VERTICAL_CENTER:
					elyt.x = cx - elyt.width/2;
					break;

				default:
					MsgOut.println("Illegal allignment option");
					break;
				}
				e.setDiagramBounds(elyt);		
			}
		}

		m_diagram.rescaleDiagram();
		repaintDg();
	}

	protected void sameSize(int sizeDim) 
	{
		Vector grp = startGroupOp();

		if (grp == null)
			return;

		// Size group based on size of key entity

		EntityInstance ke = m_diagram.getKeyEntity();
		Rectangle klyt = ke.getDiagramBounds();

		Enumeration en = grp.elements();

		while (en.hasMoreElements()) {
			EntityInstance e = (EntityInstance) en.nextElement();

			if (e != ke) {
				Rectangle elyt = (Rectangle) e.getDiagramBounds().clone();

				switch(sizeDim) {
				case /* ALT */ Do.SZ_WIDTH:
					elyt.width = klyt.width;
					break; 

				case /* ALT */ Do.SZ_HEIGHT:
					elyt.height = klyt.height;
					break;

				case /* ALT */ Do.SZ_WIDTH_HEIGHT:
					elyt.width = klyt.width;
					elyt.height = klyt.height;
					break;
				}
				e.setDiagramBounds(elyt);
			}
		}

		m_diagram.rescaleDiagram();
		repaintDg();
	} 

	protected void equalSpacing(int mode) 
	{
		Vector grp = startGroupOp();

		if (grp == null) {
			return;
		}

		if (grp.size() < 3) {
			error("Minimum group of three required");
			return;
		}

		Rectangle bb = m_diagram.getGroupBoundingBox();

		if (bb == null) {
			error("Group not selected");
			return;
		}

		// Sort entities based on vertical or horizontal position

		sortVectorByLayout(grp, mode);

		// Calculate sep

		double tw, th;
		double sep = 0.0;

		Enumeration en = grp.elements();

		tw = th = 0;

		while(en.hasMoreElements()) {
			EntityInstance e = (EntityInstance) en.nextElement();

			Rectangle lyt = e.getDiagramBounds();

			tw += lyt.width;
			th += lyt.height;
		}

		if (mode == Do.SPC_HORIZ) {
			sep = (bb.width - tw)/(grp.size()-1);
		}
		else {
			sep = (bb.height - th)/(grp.size()-1);
		}

		// Apply new layouts to second through penultimate entitites

		EntityInstance ce = (EntityInstance) grp.elementAt(0);
		Rectangle clyt = ce.getDiagramBounds();

		for (int i = 1; i < grp.size()-1; i++) {
			EntityInstance e = (EntityInstance) grp.elementAt(i);

			Rectangle lyt = e.getDiagramBounds();

			if (mode == Do.SPC_HORIZ) {
				lyt.x = (int) (clyt.x + clyt.width + sep);
			}
			else {
				lyt.y = (int) (clyt.y + clyt.height + sep);
			}

			clyt = (Rectangle) lyt.clone();
			e.setDiagramBounds(lyt);
		}

		m_diagram.rescaleDiagram();
		repaintDg();
	}

	protected void createContainedGroup() 
	{
		Vector grp = startGroupOp();

		if (grp == null) {
			return;
		}

		EntityInstance pe = ((EntityInstance) grp.firstElement()).getContainedBy();

		Rectangle nlyt = m_diagram.getGroupBoundingBox();

		nlyt.y -= RelationInstance.CLIENT_SUPPLIER_EL_LEN+5;
		nlyt.x -= 5;
		nlyt.width += 10;
		nlyt.height += RelationInstance.CLIENT_SUPPLIER_EL_LEN+10;

		String cname;
		int n = 0;

		do {
			cname = "Container#" + n;
			n++;
		} while(m_diagram.entityExists(cname));

		EntityInstance ne = m_diagram.addEntity(cname, nlyt, pe);

		Enumeration en = grp.elements();

		while (en.hasMoreElements()) {
			EntityInstance e = (EntityInstance) en.nextElement();

			Rectangle lyt = e.getDiagramBounds();

			e.moveEntityContainment(ne);

			e.setDiagramBounds(lyt);		
		}

		m_diagram.rescaleDiagram();
		repaintDg();
	}
		
	protected Vector clipboard = null;

	protected void cutGroup() 
	{
		Enumeration en;

		clipboard = m_diagram.getGroup();

		if (clipboard == null) {
			error("Group not selected");
		} else {
			// Don't want a subsequent paste to try and paste X into X
			m_diagram.clearKeyEntity();
			doFeedback("Group copied to clipboard");
		
			for (en = clipboard.elements(); en.hasMoreElements(); ) {
				EntityInstance e = (EntityInstance) en.nextElement();
				m_diagram.removeEntity(e);
		}	}
		redrawDg();
	}

	protected void pasteGroup(Object object) 
	{
		if (clipboard == null || clipboard.size() == 0) {
			error("Clipboard empty");
		} else {

			EntityInstance pe = m_diagram.targetEntity(object);
			if (pe != null) {
				Enumeration en;
				EntityInstance e;
				int		errs   = 0;
				int		cnt    = 0;

				for (en = clipboard.elements(); en.hasMoreElements(); ) {
					e = (EntityInstance) en.nextElement();
					if (m_diagram != e.getDiagram()) {
						++errs;
						error("Attempting to paste entity into a different diagram");
						continue;
					}

					if (m_diagram.containsEntity(e)) {
						// Make sure no item in clipboard isn't already in destination diagram.
						++errs;
						error("Landscape already contains some items in clipboard");
						continue;
					}
//					System.out.println("pasteGroup: " + e + " into " + pe);
					pe.addContainment(e);
					++cnt;
				}
				if (cnt != 0) {
					doFeedback("Pasted " + cnt + " entities into " + pe.getLabel());
					redrawDg();
		}	}	}
		clipboard = null;
	}

	protected void saveStatus(String stat) 
	{
		doFeedback("Save status: " + stat);
	}


	// Save the landscape in the URL

	protected String doSaveByURL(String lsSaveURL) 
	{
		URL lsURL;

		try {
			lsURL = new URL(lsSaveURL);
		}
		catch(MalformedURLException e) {
		   return "Malformed URL on write";
		}

		URLConnection urlCon;

		try {
			urlCon = lsURL.openConnection();

			urlCon.setDoInput(false);
			urlCon.setDoOutput(true);
			urlCon.setAllowUserInteraction(false);
		}
		catch (IOException e) {
			return "Couldn't open connection";
		}

		OutputStream os;

		try {
			os = urlCon.getOutputStream();
		}
		catch (IOException e) {
			return "Couldn't openStream";
		}

		try {
			m_diagram.saveDiagram(os, false);
		}
		catch (IOException e) {
			return "IOException on write";
		}

		return null;
	}

	public void saveByURL(String overridePath) 
	{
		MsgOut.dprintln("Save by URL");

		URL lsURL;

		String str = doSaveByURL(lsSavePath);

		if (str != null) {
			saveStatus(str);
			return;
		}

		if (lsSaveCmd == null) {
			saveStatus("Success");
			return;
		}

		try {
			lsURL = new URL(lsSaveCmd);
		}
		catch(MalformedURLException e) {
		   saveStatus("Malformed URL on write");
		   return;
		}

		InputStream is;
		try {
			is = lsURL.openStream();
		}
		catch (IOException e) {
			saveStatus("Couldn't openStream for store");
			return;
		}

		try {
			DataInputStream dis = new DataInputStream(is);

			str = dis.readLine();

			is.close();

			saveStatus(str);
		}
		catch(IOException e) {
			saveStatus("IOexception on store");
		}
	}

	protected String saveByFile(String newPath) 
	{
		try {
			File file = (File) m_diagram.getContext();

			if (newPath == null) {
				String name = file.getPath();

				file.renameTo(new File(name + ".old"));
			}
			else {
				// New filename
				file = new File(newPath);
				m_diagram.setContext(file);
			}

			FileOutputStream os = new FileOutputStream(file);

			m_diagram.saveDiagram(os, false);
			return null;
		}
		catch(IOException e) {
			return "IOException on file open";
		}
	}

	protected String saveByFileRaw(String newPath) 
	{
		try {
			File file = (File) m_diagram.getContext();

			if (newPath == null) {
				String name = file.getPath();

				if (name.endsWith(".raw"))
					file.renameTo(new File(name + ".old"));
				else
					file = new File(name + ".raw");
			}
			else {
				// New filename
				file = new File(newPath + ".raw");
				m_diagram.setContext(file);
			}

			FileOutputStream os = new FileOutputStream(file);

			m_diagram.saveDiagramRaw(os);
			return null;
		}
		catch(IOException e) {
			return "IOException on file open";
		}
	}

	protected String saveWithCmd(String saveSuffix, String cmdTemplate)
	{
		File file = (File) m_diagram.getContext();

		String path = file.getPath() + saveSuffix;

		File nfile = new File(path);

		// Save away the monolithic TA file

		try {

			FileOutputStream os = new FileOutputStream(nfile);

			m_diagram.saveDiagram(os, true);
		}
		catch(IOException e) {
			return "IOException on file save";
		}
	
		// Run command after save

		String cmd = Util.expand(cmdTemplate, file.getName(), this);

		// System.out.println("Save command: " + cmd);

		String rc;

		try {
			Runtime rt = Runtime.getRuntime();

			Process p = rt.exec(cmd);

			DataInputStream es = 
				new DataInputStream(p.getErrorStream());

			rc = es.readLine();

			if (rc.equals("Done"))
				rc = "Save succeeded";

			p.destroy();
		}
		catch (Exception e) {
			rc = "Exec failed for: " + cmd;
		}

		return rc;
	}

	protected void doSaveLs(String overridePath) 
	{
		doFeedback("Saving landscape...");

		if (m_diagram.getContext() instanceof URL) {
			saveByURL(overridePath);
		}
		else {
			String rc;

			if (overridePath == null && lsSaveCmd != null) {
				MsgOut.dprintln("Save by app");
				rc = saveWithCmd(lsSaveSuffix, lsSaveCmd);
			}
			else {
				MsgOut.dprintln("Save by file");
				rc = saveByFile(overridePath); 
			}

			if (rc == null)
				rc = "Success";

			saveStatus(rc);
		}
	}

	protected void saveLs() 
	{
		Object context = m_diagram.getContext();
		String txt = null;

		if (context == null) {

			txt = filePrompt("Save Landscape", m_diagram.getAbsolutePath(), FA_SAVE);
			if (txt.length() == 0) {
						return;
				}
		}

		doSaveLs(null);
	}

	protected void doSaveLsRaw(String overridePath) 
	{
		doFeedback("Saving landscape in raw format...");

		if (m_diagram.getContext() instanceof URL) {
			// saveByURL(overridePath);
		}
		else {
			MsgOut.dprintln("Save by file");

			String rc = saveByFileRaw(overridePath); 

			if (rc == null)
				rc = "Success";

			saveStatus(rc);
		}
	}

	protected void saveLsRaw() 
	{
		Object context = m_diagram.getContext();
		String txt = null;

		if (context == null) {
			txt = filePrompt("Save Landscape Raw", m_diagram.getAbsolutePath(), FA_SAVE);

			if (txt.length() == 0)
				return;
		}

		doSaveLsRaw(null);
	}

	private static boolean _printing   = false;
	private static boolean _web		   = false;
	private static boolean _landscape  = false;
	private static boolean _blackwhite = false;

	private FileOutputStream pos;

	public static boolean isPrinting() 
	{
		return _printing;
	}

	/**
	 * Create a printer graphics (for paper) using PSGr.
	 */
	private PSGr getPaperGraphics( String psPath ) throws IOException 
	{
		Rectangle dr;	// diagram rectangle
		Rectangle pr;	// print rectangle
		Rectangle bb;	// bounding box
		float	  xScale;
		float	  yScale;
		float	  xyScale;
		PSGr	  g;

		dr = m_diagram.getBounds();
							
		// scale the printing to fit on a page;

		float wPage = PSGr.PAGEWIDTH - PSGr.PAGEMARGIN*2;
		float hPage = PSGr.PAGEHEIGHT - PSGr.PAGEMARGIN*2;

		if (_landscape) {
			xScale = hPage / (float)dr.width;
			yScale = wPage / (float)dr.height;
		}
		else {
			xScale = wPage / (float)dr.width;
			yScale = hPage / (float)dr.height;
		}

		xyScale = yScale;
	
		if (xScale < yScale) {
			xyScale = xScale;
		}

		if (xyScale > 1) {
			xyScale = 1;
		}

		pos = new FileOutputStream(psPath);

		g = new PSGr( pos, getGraphics());

		// Center it in page

		if (_landscape) {
			int mx = (int) (wPage - dr.height*xyScale)/2;
			int my = (int) (hPage - dr.width*xyScale)/2;

			int tx = (int) -(dr.x * xyScale);
			int ty = (int) -(dr.y * xyScale);

			// g.translate(tx, ty);
			g.translate(-71, 100);

			g.setBoundingBox( new Rectangle(mx, my, 
							(int)(dr.height * xyScale + 3), 
							(int)(dr.width * xyScale + 2) ) );
		}
		else {
			int mx = (int) (wPage - dr.width*xyScale)/2;
			int my = (int) (hPage - dr.height*xyScale)/2;

			int tx = (int) -(dr.x * xyScale) + mx;
			int ty = (int) -(dr.y * xyScale) + my;

			g.translate(tx, ty);

			g.setBoundingBox( new Rectangle(mx, my,
							(int)(dr.width * xyScale + 2),
							(int)(dr.height * xyScale + 3) ) );
		}

		g.setModes(_landscape, _blackwhite, -1);
		g.scale( xyScale, xyScale );

		return g;
	}

	private PSGr getBrowserGraphics( String psPath ) throws IOException 
	{
		Rectangle dr;	// diagram rectangle
		Rectangle pr;	// print rectangle
		Rectangle bb;	// bounding box
		float	  xScale;
		float	  yScale;
		float	  xyScale;
		PSGr	  g;

		dr = m_diagram.getBounds();
		pr = new Rectangle(0, 0, PSGr.PAGEWIDTH, PSGr.PAGEHEIGHT);
							
		// scale the printing to fit on a page;
		// based on PSGr.scalePaint()

		xScale = pr.width / (float)dr.width;
		yScale = pr.height / (float)dr.height;
		xyScale = yScale;
	
		if (xScale < yScale) {
			xyScale = xScale;
		}

		if (xyScale > 1) {
			xyScale = 1;
		}

		pos = new FileOutputStream(psPath);

		g = new PSGr( pos, getGraphics(),
						0, //PSGr.PAGEMARGIN,
						PSGr.PAGEWIDTH, PSGr.PAGEHEIGHT);

		g.setModes(_landscape, _blackwhite, (int) (dr.height * xyScale - 4));

		// Push it to the top of page, with no margins

		int tx = (int) -(dr.x * xyScale);
		int ty = (int) -(dr.y * xyScale);

		g.translate(tx, ty);

		g.setBoundingBox( new Rectangle(0, 0,
							(int)(dr.width * xyScale + 2),
							(int)(dr.height * xyScale + 3) ) );

		g.scale( xyScale, xyScale );
		return g;
	}

	/**
	 * Print the diagram to a postscript file using the PSGr class
	 * PSGr is (C) 1996 E.J. Friedman-Hill and Sandia National Labs.
	 */
	protected boolean doPrint(String psPath) 
	{
		PSGr	  g;
		int		  dx, dy;
		float	  scale;

		boolean success = true;

		BaseEntity.set3Dlook(!_blackwhite);

		try {
			_printing = true;

			if (_web) {
				g = getBrowserGraphics(psPath);
			} else {
				g = getPaperGraphics(psPath);
			}
			m_diagram.getDrawRoot().validate();
			g.showpage();
			g.dispose();
			g.writeTrailer();

		} catch( Exception e ) {
			System.err.println("Printing..oops: " + e );
			e.printStackTrace();
			success = false;
		} finally {
			_printing = false;
		}

		try {
			pos.close();
		}
		catch( Exception e) {
			System.err.println("Printing..oops: " + e );
			success = false;
		}

		BaseEntity.set3Dlook(true);

		return success;
	}

	protected void printLs(String psPath) 
	{
		doFeedback("Printing landscape to: " + psPath );

		boolean rc = doPrint(psPath);

		if (rc)
			doFeedback("Done printing (success)" );
		else
			error("Done printing (failed)" );
	}

	protected String getLsPath(EntityInstance e) 
	{

		if (e == m_diagram.getRootInstance()) {
			return Util.mungeName(e.getLabel());
		}

		Vector v = new Vector();

		EntityInstance ce = e;

		String path = "";

		do {
			v.addElement(Util.mungeName(ce.getLabel()));
			do {
				ce = ce.getContainedBy();			
			} while (ce != null && !ce.isEnterable());

		} while (ce != null);

		for (int i = v.size()-2; i >= 0; i--) {
			String name = (String) v.elementAt(i);

			path += name;

			if (i > 0) {
				path += "#";
			}
		}

		return path;
	}

	protected EntityInstance printAllLs(String dir, EntityInstance e, int minNumChildren)
	{
		if (e.numChildren() < minNumChildren)
			return null;

		if (e.isEnterable()) {
			m_diagram.setPreserveEntityMarks(EntityInstance.PRESENTATION_MARKS);
			m_diagram.setPreserveRelationMarks(RelationInstance.PRESENTATION_MARKS);
			m_diagram.navigateTo(e);

			String psPath = dir + File.separatorChar + getLsPath(m_diagram.getDrawRoot()) + ".eps";

			doFeedback("Printing: " + e.getLabel() );

			if (!doPrint(psPath))
				return e;
		}

		Enumeration en = e.getChildren();

		while (en.hasMoreElements()) {
			EntityInstance ce = (EntityInstance) en.nextElement();

			EntityInstance fe = printAllLs(dir, ce, minNumChildren);

			if (fe != null)
				return fe;
		}

		return null;
	}

	protected void printAllLs(String dir) 
	{
		EntityInstance oldRoot = m_diagram.getDrawRoot();

		EntityInstance top = m_diagram.getRootInstance();

		EntityInstance fe = printAllLs(dir, top, 2);

		if (fe != null) {
			error("Print to landscape failed on: " + fe.getLabel());
		}
		else {
			doFeedback("Done printing (success)" );
		}

		m_diagram.setPreserveEntityMarks(EntityInstance.PRESENTATION_MARKS);
		m_diagram.setPreserveRelationMarks(RelationInstance.PRESENTATION_MARKS);
		m_diagram.navigateTo(oldRoot);
	}

	protected void printBook(String psPath) 
	{
	}

	public void dialogAction(Object[] results, int mode) 
	{
		String txt = (String) results[0];

		if (txt.length() != 0) {
			switch(mode)
			{
			case FA_SAVE:
				doSaveLs(txt);
				break;

			case FA_LOAD:
				loadLs(txt);
				break;

			case FA_PRINT:
				_web = ((Boolean) results[1]).booleanValue();
				_landscape = ((Boolean) results[2]).booleanValue();
				_blackwhite = ((Boolean) results[3]).booleanValue();

				printLs(txt);
				break;

			case FA_PRINT_ALL:
				_web = ((Boolean) results[1]).booleanValue();
				_landscape = ((Boolean) results[2]).booleanValue();
				_blackwhite = ((Boolean) results[3]).booleanValue();

				printAllLs(txt);
				break;
			}
		}
	}
	
	protected void initialLoad() 
	{
		super.initialLoad();

		if (lsPath != null) {
			diagrams.addElement(m_diagram);
			diagramNames.addElement(lsPath);
			registerLsName(m_diagram.getName());
		} else {
			diagrams.addElement(m_diagram);
			diagramNames.addElement("");
			registerLsName("unnamed");
		}
	}

	public boolean isViewer() 
	{
		return false;
	}

	// --------------
	// Event handling 
	// --------------

	// Overrides LandscapeViewerCore.processKey() 

	public void processKey(int key, int modifiers, Object object) 
	{
		String str; 

		if ((modifiers & Event.CTRL_MASK) != 0) {
			switch(key) {
			case Do.OPEN_LANDSCAPE:
			{
				Object context = m_diagram.getContext();
				String txt;

				if (context != null) {
					txt = filePrompt( "Load Landscape", m_diagram.getAbsolutePath(), FA_LOAD);
				} else {
					txt = filePrompt("Load Landscape", "", FA_LOAD);
				}
				loadLs(txt);
				return;
			}
			case Do.PRINT_LANDSCAPE:
			{
				Object context = m_diagram.getContext();
				String path, txt;

				String title = "Print Landscape Page to PostScript .eps File";

				if (context != null) {
					path = m_diagram.getDir() + File.separatorChar + Util.mungeName(m_diagram.getDrawRoot().getLabel() + ".eps");
				} else {
					path = Util.mungeName(m_diagram.getDrawRoot().getLabel() + ".eps");
				}

				txt = filePrompt("Save to eps file:", path, FA_SAVE);
//				txt = printPrompt(title, "Print to file:", path, FA_PRINT);
				if (txt == null || txt.length() == 0) {
					return;
				}
				printLs(txt);
				return;
			}
			case Do.SAVE_AS:
			{
				Object context = m_diagram.getContext();
				String txt;

				if (context != null) {
					txt = filePrompt("Save Landscape As", m_diagram.getAbsolutePath(), FA_SAVE);
				} else {
					txt = filePrompt("Save Landscape As", "", FA_SAVE);
				}
				if (txt == null || txt.length() == 0) {
					return;
				}
				doSaveLs(txt);
				return;
			}
			case Do.SAVE:
				if (m_diagram.isReadOnly()) {
					error("Diagram is read-only");
					return;
				}
				if ((modifiers & Event.ALT_MASK) != 0) {
					saveLsRaw();
				} else {
					saveLs();
				}
				return;
			case Do.PASTE:
				pasteGroup(object);
				return;
			case Do.CUT:
				cutGroup();
				return;
			}
		} else if ((modifiers & Event.ALT_MASK) != 0) {
			switch (key) {
			case Do.A_HORIZ_TOP:
			case Do.A_HORIZ_CENTER:
			case Do.A_VERTICAL_LEFT:
			case Do.A_VERTICAL_RIGHT:
			case Do.A_VERTICAL_CENTER:
				align(key);
				return;
			case Do.A_FIT_LABEL:
				fitToLabel();
				return;
			case Do.A_FIT:
				fitChildren();
				return;
			case Do.A_GROUP:
				createContainedGroup();
				return;
			case Do.SZ_WIDTH:
			case Do.SZ_HEIGHT:
			case Do.SZ_WIDTH_HEIGHT:
				sameSize(key);
				return;
			case Do.SPC_HORIZ:
			case Do.SPC_VERTICAL:
				equalSpacing(key);
				return;
			case Do.ARROW_DIMENSIONS:
				ArrowDimensions.create(this);
				repaint();
				return;
			}

		} else if (modifiers == 0) { 
			switch (key) {
			case Do.VIEW_MODE:
			{
				if (m_modeHandler == m_editModeHandler) {
					doFeedback("Set to viewer mode");
					m_modeHandler = m_viewModeHandler;
					m_diagram.setDrawBends(false);
					setViewMode(true);
				} else {
					doFeedback("Set to editor mode");
					m_modeHandler = m_editModeHandler;
					m_diagram.setDrawBends(true);
					setViewMode(false);
				}
				m_diagram.setModeHandler(m_modeHandler);

				repaintDg();
				return;
			}
			case Do.GRID_INCREASE:
			{
				int grid = m_diagram.getGrid();

				if (grid == GRID_MAX) {
					error("Grid already set to maximum (" + grid + " pixels)");
				} else {
					int ng = grid+1;

					m_diagram.setGrid(ng);
					doFeedback("Grid set to " + ng + " pixels");
				}
				return;
			}
			case Do.GRID_DECREASE:
			{
				int grid = m_diagram.getGrid();

				if (grid == 1) {
					error("Grid already set to minimum (1 pixel)");
				} else {
					int ng = grid-1;

					m_diagram.setGrid(ng);
					doFeedback("Grid set to " + ng + " pixel(s)");
				}
				return;
			}}
		}
		super.processKey(key, modifiers, object);
	}

	// Overrides LandscapeViewer.isReadWrite()
	public boolean isreadWrite() {
		// Eventually this should be changed to return true
		// iff the landscape is definitely writeable
		// For now, assume true in editor

		return true;
	}

	public InternalBufferStream getInternalBufferStream() {
		return null;
	}

	// ItemListener interface

	public void itemStateChanged(ItemEvent ev)
	{
		int ind = m_lsDropDown.getSelectedIndex();
		setDiagram(ind);
		requestFocus();

	}
} 
